from enum import Enum

import multiprocessing.pool
import typing

from paperboat.model_loader import ModelRetriever
from paperboat.processing import Step, CarryOnData
from segmentation.postprocessing.layout_processing import process_layout, AnalyzedContent
from segmentation.postprocessing.layout_settings import LayoutProcessingSettings
from segmentation.predictors import Predictor, PredictionSettings
from segmentation.preprocessing.source_image import SourceImage


class BaselinePrediction(Step):
    def __init__(self, predictor: Predictor):
        self.nn_predictor = predictor

    @classmethod
    def get_id(cls) -> str:
        return "segmentation"

    def run(self, data: CarryOnData) -> CarryOnData:
        #get the source image
        source_image = data.get("source_image")

        prediction, scaled_image = self.nn_predictor.predict_image(source_image)
        data.set("baseline.prediction", prediction)
        data.set("baseline.scaled_image",scaled_image)
        return data

    def input_format(self) -> typing.Any:
        return CarryOnData

    def output_format(self) -> typing.Any:
        return CarryOnData

    @classmethod
    def from_config(cls, config: dict) -> 'BaselinePrediction':
        # we have to load the model here
        # and get the prediction settings
        if config.get("model_path") is not None:
            model_path = config.get("model_path")
        elif config.get("repo_model") is not None:
            model_path = ModelRetriever({}).get_lib().get(config.get("repo_model"))
        else:
            model_path = ModelRetriever({}).get_lib().get_segmentation_model()

        settings = PredictionSettings([model_path],config.get("scale_area"), config.get("min_line_height"), config.get("max_line_height"))

        # make the predictor
        predictor = Predictor(settings)
        return BaselinePrediction(predictor=predictor)


class SegmentationLayoutAnalyser(Step):
    def __init__(self, layout_settings: LayoutProcessingSettings):
        self.layout_settings = layout_settings

    @classmethod
    def get_id(cls) -> str:
        return "layout"

    def run(self, data: CarryOnData) -> CarryOnData:
        prediction, scaled_image = data.get("baseline.prediction"), data.get("baseline.scaled_image")

        scale_factor = prediction.prediction_scale_factor
        analyzed_content = process_layout(prediction,
                                          scaled_image,
                                          multiprocessing.pool.ThreadPool(),
                                          self.layout_settings)

        # layout_debugging(args, analyzed_content, scaled_image, file)

        # convert this back to the original image space
        data.set("layout.analyzed_content.prediction_space",analyzed_content)
        data.set("layout.analyzed_content",analyzed_content.to_pagexml_space(scale_factor))
        return data

    def input_format(self) -> typing.Any:
        return CarryOnData

    def output_format(self) -> typing.Any:
        return CarryOnData

    @classmethod
    def from_config(cls, config: dict) -> 'Step':
        # TODO: implement configuration reading here
        return SegmentationLayoutAnalyser(LayoutProcessingSettings())

class SegmentationPageXMLOutputMode(Enum):
    DEFAULT = "default"

class SegmentationPageXMLGenerator(Step):
    def __init__(self, output_mode: SegmentationPageXMLOutputMode):
        self.mode = output_mode

    @classmethod
    def get_id(cls) -> str:
        return "segmentation_pxml_gen"

    def run(self, data: CarryOnData) -> CarryOnData:
        assert data.has("layout.analyzed_content")
        analyzed_content: AnalyzedContent = data.get("layout.analyzed_content")
        source_image: SourceImage = data.get("source_image")

        xml_gen = analyzed_content.export(source_image, data.get("src.filename"), simplified_xml=self.mode != SegmentationPageXMLOutputMode.DEFAULT)
        xml_str = xml_gen.baselines_to_xml_string()
        data.set("layout.pxml", xml_str)
        return data

    def input_format(self) -> typing.Any:
        return CarryOnData

    def output_format(self) -> typing.Any:
        return CarryOnData

    @classmethod
    def from_config(cls, config: dict) -> 'Step':
        return SegmentationPageXMLGenerator(
            output_mode=SegmentationPageXMLOutputMode(
                config.get("mode","default")))